package ru.dbotthepony.kstarbound.util

fun <T> KOptional(value: T) = KOptional.of(value)

/**
 * [java.util.Optional] supporting nulls
 *
 * This is done for structures, where value can be absent,
 * or can be present, including literal "null" as possible present value,
 * in more elegant solution than handling nullable Optionals
 */
class KOptional<T> private constructor(private val _value: T, val isPresent: Boolean) {
	val value: T get() {
		if (isPresent) {
			return _value
		}

		throw NoSuchElementException("No value is present")
	}

	inline fun ifPresent(block: (T) -> Unit) {
		if (isPresent) {
			block.invoke(value)
		}
	}

	override fun equals(other: Any?): Boolean {
		return this === other || other is KOptional<*> && isPresent == other.isPresent && _value == other._value
	}

	override fun hashCode(): Int {
		return _value.hashCode()
	}

	override fun toString(): String {
		if (isPresent) {
			return "KOptional[value = $value]"
		} else {
			return "KOptional[empty]"
		}
	}

	companion object {
		private val EMPTY = KOptional(null, false)
		private val NULL = KOptional(null, true)

		@JvmStatic
		fun <T> empty(): KOptional<T> {
			return EMPTY as KOptional<T>
		}

		@JvmStatic
		fun <T> of(value: T): KOptional<T> {
			if (value == null) {
				return NULL as KOptional<T>
			} else {
				return KOptional(value, true)
			}
		}
	}
}
